==> qtools_profiler-2.2.x-dev.tar.gz: qtools_profiler/modules/qtools_cache_profiler/src/Renderer.php <==
<?php

namespace Drupal\qtools_cache_profiler;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Render\Markup;
use Drupal\Core\Render\Renderer as CoreRenderer;
use Drupal\qtools_cache_profiler\Cache\CacheProfilerActive;
use Drupal\qtools_common\QToolsCacheHelper;

/**
 * Decorates core's Renderer to provide the necessary metadata to JS cache inspector.
 *
 * Based on renderviz module's implementation.
 */
class Renderer extends CoreRenderer {

  /**
   * {@inheritdoc}
   */
  protected function doRender(&$elements, $is_root_call = FALSE) {
    // Always add profiler context to output.
    $elements['#cache']['contexts'][] = CacheProfilerActive::CACHE_CONTEXT_KEY;

    // If profiler is not active - skip any additional processing.
    if (!$this->profilingIsActive()) {
      return parent::doRender($elements, $is_root_call);
    }

    // Initial lazy builder info.
    $lazy_builder = $elements['#lazy_builder'] ?? NULL;
    $lazy_callback_key = !empty($lazy_builder) ?
      $elements['#lazy_builder'][0] . ':' . ($original_elements['#lazy_builder'][1][0] ?? '') :
      NULL;
    $create_placeholder = $elements['#create_placeholder'] ?? NULL;

    $original_elements = $elements;
    $original_elements['#cache'] = $elements['#cache'] ?? [];

    // Do normal rendering.
    $start = microtime(TRUE);
    $result = parent::doRender($elements, $is_root_call);
    $elapsed = microtime(TRUE) - $start;

    // If item was loaded from cache get real cache id.
    $realCacheId = $elements['#qtcp_cid'] ?? NULL;

    // Also, generate "expected one".
    $cacheId = !empty($elements['#cache']['keys']) ?
      QToolsCacheHelper::buildCidElements($elements, $elements['#cache']['keys']) :
      '';

    $lazy_builder_built = $elements['#lazy_builder_built'] ?? NULL;
    $element_markup = $elements['#markup'] ?? '';
    $drupal_placeholder = str_starts_with($element_markup, '<drupal-render-placeholder callback="');
    $big_pipe_placeholder = str_starts_with($element_markup, '<span data-big-pipe-placeholder-id="');
    $lazy_loaded = !empty($lazy_builder) && !$create_placeholder;

    // @todo visualise no output.
    // When there is no output, there is also nothing to visualize.
    if ($result === '') {
      return '';
    }

    // The HTML spec says HTML comments are markup, thus it's not allowed to use
    // HTML comments inside HTML element attributes. For example:
    // <a <!--title="need to be comment out"-->>a link</a>
    // is as wrong as
    // <a <span></span>>a link</a>
    // So we only wrap $result in a HTML comment with metadata when
    // $result actually contains HTML markup. So, 'This is text.' will cause an
    // early return, but 'This is text and <a href="…">a link</a>.' will not.
    // The presence of HTML indicates it's valid to have HTML, and when it is
    // valid to have HTML, HTML comments are allowed too.
    // Since strip tags always remove comments we need to make sure
    // they will not affect our compassing.
    $cleaned_result = trim(preg_replace('/<!--(.*)-->/Uis', '', $result));

    // Check if this chunk is HTML markup.
    // Big pipe will add nojs placeholders that will not become HTML, so
    // we ignore those too (they are usually used in attributes with urls).
    $big_pipe_tags = [
      'nojs-bigpipe-placeholder-styles-placeholder',
      'nojs-bigpipe-placeholder-scripts-placeholder',
    ];
    if ($cleaned_result == strip_tags($cleaned_result, $big_pipe_tags)) {
      // Returned without debug output.
      return $result;
    }

    // Apply the same default cacheability logic that Renderer::doRender()
    // applies.
    $pre_bubbling_elements = $original_elements;
    $pre_bubbling_elements['#cache']['tags'] = $original_elements['#cache']['tags'] ?? [];
    $pre_bubbling_elements['#cache']['max-age'] = $original_elements['#cache']['max-age'] ?? Cache::PERMANENT;
    // @todo Add these always? That's more accurate for visualization purposes;
    //   it is only for performance optimization purposes that the wrapped
    //   function doesn't do that.
    if ($is_root_call || isset($elements['#cache']['keys'])) {
      $required_cache_contexts = $this->rendererConfig['required_cache_contexts'];
      if (isset($pre_bubbling_elements['#cache']['contexts'])) {
        $pre_bubbling_elements['#cache']['contexts'] = Cache::mergeContexts($pre_bubbling_elements['#cache']['contexts'], $required_cache_contexts);
      }
      else {
        $pre_bubbling_elements['#cache']['contexts'] = $required_cache_contexts;
      }
    }

    // Prepare data for output.
    $result_cache = $elements['#cache'];
    $pre_bubbling_cache = $pre_bubbling_elements['#cache'];

    // Small exception for global document element.
    if (strpos($cleaned_result, '<!DOCTYPE') === 0) {
      // Main page element.
      $result_cache['contexts'][] = 'route';
      $result_cache['contexts'][] = 'request_format';
      $result_cache['contexts'][] = 'url.query_args:_wrapper_format';
      $cacheId = QToolsCacheHelper::buildCidElements(['#cache' => $result_cache], ['_response_']);
    }

    // Add debug output.
    // @todo This currently prints the final and pre-bubbling elements.
    $interesting_keys = ['keys', 'contexts', 'tags', 'max-age'];
    if (!array_intersect(array_keys($result_cache), $interesting_keys) &&
      !array_intersect(array_keys($pre_bubbling_cache), $interesting_keys)
    ) {
      return $elements['#markup'];
    }

    // Add debug info and wrappers.
    if (!empty($result_cache['contexts'])) {
      $result_cache['contexts'] = array_values($result_cache['contexts']);
    }
    if (!empty($result_cache['tags'])) {
      $result_cache['tags'] = array_values($result_cache['tags']);
    }
    if (!empty($pre_bubbling_cache['contexts'])) {
      $pre_bubbling_cache['contexts'] = array_values($pre_bubbling_cache['contexts']);
    }
    if (!empty($pre_bubbling_cache['tags'])) {
      $pre_bubbling_cache['tags'] = array_values($pre_bubbling_cache['tags']);
    }

    // Keys aren't particularly interesting if we have CID
    // and just make it harder to find info.
    $cid_info = [
      'e' => round($elapsed, 3),
      'cid' => $realCacheId ?? $cacheId,
      'keys' => $elements['#cache']['keys'] ?? NULL,
      'pre_keys' => $original_elements['#cache']['keys'] ?? NULL,
    ];
    if ($lazy_loaded) {
      $cid_info['lazy'] = TRUE;
    }

    // Order elements for ease of inspection.
    $result_cache = [
      'max-age' => $result_cache['max-age'] ?? '',
      'contexts' => $result_cache['contexts'] ?? [],
      'tags' => $result_cache['tags'] ?? [],
    ];
    $pre_bubbling_cache = [
      'max-age' => $pre_bubbling_cache['max-age'] ?? '',
      'contexts' => $pre_bubbling_cache['contexts'] ?? [],
      'tags' => $pre_bubbling_cache['tags'] ?? [],
    ];

    $page_prefix = &drupal_static('qtools_profiler_cache_page_prefix', substr(uniqid(), -2));
    $element_id = &drupal_static('qtools_profiler_cache_element_id', 0);
    $element_id++;
    $element_uuid = $page_prefix . '_' . $element_id;
    if ($lazy_loaded) {
      $element_uuid .= '_LL';
    }

    $prefix = PHP_EOL . PHP_EOL . '<!--RENDERER_START: ' . $element_uuid . '-->' . PHP_EOL . '<!--' . Json::encode($pre_bubbling_cache) . '-->' . PHP_EOL . '<!--' . Json::encode($result_cache) . '-->';
    $suffix = PHP_EOL . '<!--RENDERER_END: ' . $element_uuid . '-->';

    // If this render support caching add cid for easy visualise.
    $prefix .= PHP_EOL. '<!--' . Json::encode($cid_info) . '-->';

    // Placeholder elements always wrapped as their cache info is yet unknown.
    if (!empty($drupal_placeholder)) {
      $elements['#markup'] = PHP_EOL . '<div class="qtcp-d-placeholder">' . $elements['#markup'] . '</div>';
    }
    elseif ($big_pipe_placeholder) {
      $elements['#markup'] = PHP_EOL . '<div class="qtcp-bp-placeholder">' . $elements['#markup'] . '</div>';
    }
    elseif (!empty($cacheId) && $wrap = $this->requireWrap($elements, $cleaned_result)) {
      // Some templates do not have any additional tags, therefor
      // there is nothing to attach our data to, as we add wrapper if allowed
      // and this data is critical.
      $elements['#markup'] = PHP_EOL . '<div class="qtcp-extra-wrapper" data-qtcp-wrap="' . $wrap . '">' . $elements['#markup'] . '</div>';
    }
    $elements['#markup'] = Markup::create($prefix . $elements['#markup'] . $suffix);

    return $elements['#markup'];
  }

  /**
   * Check if given content require wraping to correcly place metadata.
   */
  protected function requireWrap($elements, $cleaned_content) {
    $trimmed_content = trim($elements['#markup']);

    // To preserve all cache info we need a target tag, so we have to wrap
    // output but we can't wrap some special tags or drupal will break.
    $unwrappable_tags = [
      '<script',
      '<link',
      '<style',
      '<meta',
      '<title',
      '<body',
      '<html',
      '<head',
      '<!DOCTYPE',
    ];
    foreach ($unwrappable_tags as $tag) {
      if (strpos($cleaned_content, $tag) === 0) {
        return FALSE;
      }
    }

    // If template is empty we would see previous render info.
    $wrap = strpos($trimmed_content, '<!--RENDERER_START-->') === 0 ?
      1 :
      FALSE;

    // If elements are about to be lazy built we wrap them.
    if (empty($wrap) && !empty($elements['#lazy_builder'])) {
      $wrap = 2;
    }

    // If there are more than one child element we need to wrap.
    if (empty($wrap)) {
      $dom = new \DOMDocument();

      libxml_use_internal_errors(TRUE);
      $dom->loadHTML($cleaned_content);

      // Structure will be wrapped in <html><body> so we go 2 level deep.
      $childNodes = $dom->childNodes[1]->childNodes[0]->childNodes;
      if ($childNodes->length > 1) {
        $wrap = urlencode($cleaned_content);
      }
      libxml_use_internal_errors(FALSE);
    }

    return $wrap;
  }

  /**
   * For profiler to be active special cookie is required.
   */
  protected function profilingIsActive(): bool {
    /** @var \Drupal\qtools_cache_profiler\Cache\CacheProfilerActive $profilerCacheContext */
    $profilerCacheContext = \Drupal::service('cache_context.qtools_cache_profiler_active');
    return $profilerCacheContext->getContext();
  }

}

==> qtools_profiler-2.3.x-dev.tar.gz: qtools_profiler/modules/qtools_cache_profiler/src/Renderer.php <==
<?php

namespace Drupal\qtools_cache_profiler;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Render\Markup;
use Drupal\Core\Render\Renderer as CoreRenderer;
use Drupal\qtools_cache_profiler\Cache\CacheProfilerActive;
use Drupal\qtools_common\QtoolsCacheHelper;

/**
 * Decorates core's Renderer to provide the necessary metadata to JS cache inspector.
 *
 * Based on renderviz module's implementation.
 */
class Renderer extends CoreRenderer {

  /**
   * {@inheritdoc}
   */
  protected function doRender(&$elements, $is_root_call = FALSE) {
    // Always add profiler context to output.
    $elements['#cache']['contexts'][] = CacheProfilerActive::CACHE_CONTEXT_KEY;

    // If profiler is not active - skip any additional processing.
    if (!$this->profilingIsActive()) {
      return parent::doRender($elements, $is_root_call);
    }

    // Initial lazy builder info.
    $lazy_builder = $elements['#lazy_builder'] ?? NULL;
    $lazy_callback_key = !empty($lazy_builder) ?
      $elements['#lazy_builder'][0] . ':' . ($original_elements['#lazy_builder'][1][0] ?? '') :
      NULL;
    $create_placeholder = $elements['#create_placeholder'] ?? NULL;

    $original_elements = $elements;
    $original_elements['#cache'] = $elements['#cache'] ?? [];

    // Do normal rendering.
    $start = microtime(TRUE);
    $result = parent::doRender($elements, $is_root_call);
    $elapsed = microtime(TRUE) - $start;

    // If item was loaded from cache get real cache id.
    $realCacheId = $elements['#qtcp_cid'] ?? NULL;

    // Also, generate "expected one".
    $cacheId = !empty($elements['#cache']['keys']) ?
      QtoolsCacheHelper::buildCidElements($elements, $elements['#cache']['keys']) :
      '';

    $lazy_builder_built = $elements['#lazy_builder_built'] ?? NULL;
    $element_markup = $elements['#markup'] ?? '';
    $drupal_placeholder = str_starts_with($element_markup, '<drupal-render-placeholder callback="');
    $big_pipe_placeholder = str_starts_with($element_markup, '<span data-big-pipe-placeholder-id="');
    $lazy_loaded = !empty($lazy_builder) && !$create_placeholder;

    // @todo visualise no output.
    // When there is no output, there is also nothing to visualize.
    if ($result === '') {
      return '';
    }

    // The HTML spec says HTML comments are markup, thus it's not allowed to use
    // HTML comments inside HTML element attributes. For example:
    // <a <!--title="need to be comment out"-->>a link</a>
    // is as wrong as
    // <a <span></span>>a link</a>
    // So we only wrap $result in a HTML comment with metadata when
    // $result actually contains HTML markup. So, 'This is text.' will cause an
    // early return, but 'This is text and <a href="…">a link</a>.' will not.
    // The presence of HTML indicates it's valid to have HTML, and when it is
    // valid to have HTML, HTML comments are allowed too.
    // Since strip tags always remove comments we need to make sure
    // they will not affect our compassing.
    $cleaned_result = trim(preg_replace('/<!--(.*)-->/Uis', '', $result));

    // Check if this chunk is HTML markup.
    // Big pipe will add nojs placeholders that will not become HTML, so
    // we ignore those too (they are usually used in attributes with urls).
    $big_pipe_tags = [
      'nojs-bigpipe-placeholder-styles-placeholder',
      'nojs-bigpipe-placeholder-scripts-placeholder',
    ];
    if ($cleaned_result == strip_tags($cleaned_result, $big_pipe_tags)) {
      // Returned without debug output.
      return $result;
    }

    // Apply the same default cacheability logic that Renderer::doRender()
    // applies.
    $pre_bubbling_elements = $original_elements;
    $pre_bubbling_elements['#cache']['tags'] = $original_elements['#cache']['tags'] ?? [];
    $pre_bubbling_elements['#cache']['max-age'] = $original_elements['#cache']['max-age'] ?? Cache::PERMANENT;
    // @todo Add these always? That's more accurate for visualization purposes;
    //   it is only for performance optimization purposes that the wrapped
    //   function doesn't do that.
    if ($is_root_call || isset($elements['#cache']['keys'])) {
      $required_cache_contexts = $this->rendererConfig['required_cache_contexts'];
      if (isset($pre_bubbling_elements['#cache']['contexts'])) {
        $pre_bubbling_elements['#cache']['contexts'] = Cache::mergeContexts($pre_bubbling_elements['#cache']['contexts'], $required_cache_contexts);
      }
      else {
        $pre_bubbling_elements['#cache']['contexts'] = $required_cache_contexts;
      }
    }

    // Prepare data for output.
    $result_cache = $elements['#cache'];
    $pre_bubbling_cache = $pre_bubbling_elements['#cache'];

    // Small exception for global document element.
    if (strpos($cleaned_result, '<!DOCTYPE') === 0) {
      // Main page element.
      $result_cache['contexts'][] = 'route';
      $result_cache['contexts'][] = 'request_format';
      $result_cache['contexts'][] = 'url.query_args:_wrapper_format';
      $cacheId = QtoolsCacheHelper::buildCidElements(['#cache' => $result_cache], ['_response_']);
    }

    // Add debug output.
    // @todo This currently prints the final and pre-bubbling elements.
    $interesting_keys = ['keys', 'contexts', 'tags', 'max-age'];
    if (!array_intersect(array_keys($result_cache), $interesting_keys) &&
      !array_intersect(array_keys($pre_bubbling_cache), $interesting_keys)
    ) {
      return $elements['#markup'];
    }

    // Add debug info and wrappers.
    if (!empty($result_cache['contexts'])) {
      $result_cache['contexts'] = array_values($result_cache['contexts']);
    }
    if (!empty($result_cache['tags'])) {
      $result_cache['tags'] = array_values($result_cache['tags']);
    }
    if (!empty($pre_bubbling_cache['contexts'])) {
      $pre_bubbling_cache['contexts'] = array_values($pre_bubbling_cache['contexts']);
    }
    if (!empty($pre_bubbling_cache['tags'])) {
      $pre_bubbling_cache['tags'] = array_values($pre_bubbling_cache['tags']);
    }

    // Keys aren't particularly interesting if we have CID
    // and just make it harder to find info.
    $cid_info = [
      'e' => round($elapsed, 3),
      'cid' => $realCacheId ?? $cacheId,
      'keys' => $elements['#cache']['keys'] ?? NULL,
      'pre_keys' => $original_elements['#cache']['keys'] ?? NULL,
    ];
    if ($lazy_loaded) {
      $cid_info['lazy'] = TRUE;
    }

    // Order elements for ease of inspection.
    $result_cache = [
      'max-age' => $result_cache['max-age'] ?? '',
      'contexts' => $result_cache['contexts'] ?? [],
      'tags' => $result_cache['tags'] ?? [],
    ];
    $pre_bubbling_cache = [
      'max-age' => $pre_bubbling_cache['max-age'] ?? '',
      'contexts' => $pre_bubbling_cache['contexts'] ?? [],
      'tags' => $pre_bubbling_cache['tags'] ?? [],
    ];

    $page_prefix = &drupal_static('qtools_profiler_cache_page_prefix', substr(uniqid(), -2));
    $element_id = &drupal_static('qtools_profiler_cache_element_id', 0);
    $element_id++;
    $element_uuid = $page_prefix . '_' . $element_id;
    if ($lazy_loaded) {
      $element_uuid .= '_LL';
    }

    $prefix = PHP_EOL . PHP_EOL . '<!--RENDERER_START: ' . $element_uuid . '-->' . PHP_EOL . '<!--' . Json::encode($pre_bubbling_cache) . '-->' . PHP_EOL . '<!--' . Json::encode($result_cache) . '-->';
    $suffix = PHP_EOL . '<!--RENDERER_END: ' . $element_uuid . '-->';

    // If this render support caching add cid for easy visualise.
    $prefix .= PHP_EOL. '<!--' . Json::encode($cid_info) . '-->';

    // Placeholder elements always wrapped as their cache info is yet unknown.
    if (!empty($drupal_placeholder)) {
      $elements['#markup'] = PHP_EOL . '<div class="qtcp-d-placeholder">' . $elements['#markup'] . '</div>';
    }
    elseif ($big_pipe_placeholder) {
      $elements['#markup'] = PHP_EOL . '<div class="qtcp-bp-placeholder">' . $elements['#markup'] . '</div>';
    }
    elseif (!empty($cacheId) && $wrap = $this->requireWrap($elements, $cleaned_result)) {
      // Some templates do not have any additional tags, therefor
      // there is nothing to attach our data to, as we add wrapper if allowed
      // and this data is critical.
      $elements['#markup'] = PHP_EOL . '<div class="qtcp-extra-wrapper" data-qtcp-wrap="' . $wrap . '">' . $elements['#markup'] . '</div>';
    }
    $elements['#markup'] = Markup::create($prefix . $elements['#markup'] . $suffix);

    return $elements['#markup'];
  }

  /**
   * Check if given content require wraping to correcly place metadata.
   */
  protected function requireWrap($elements, $cleaned_content) {
    $trimmed_content = trim($elements['#markup']);

    // To preserve all cache info we need a target tag, so we have to wrap
    // output but we can't wrap some special tags or drupal will break.
    $unwrappable_tags = [
      '<script',
      '<link',
      '<style',
      '<meta',
      '<title',
      '<body',
      '<html',
      '<head',
      '<!DOCTYPE',
    ];
    foreach ($unwrappable_tags as $tag) {
      if (strpos($cleaned_content, $tag) === 0) {
        return FALSE;
      }
    }

    // If template is empty we would see previous render info.
    $wrap = strpos($trimmed_content, '<!--RENDERER_START-->') === 0 ?
      1 :
      FALSE;

    // If elements are about to be lazy built we wrap them.
    if (empty($wrap) && !empty($elements['#lazy_builder'])) {
      $wrap = 2;
    }

    // If there are more than one child element we need to wrap.
    if (empty($wrap)) {
      $dom = new \DOMDocument();

      libxml_use_internal_errors(TRUE);
      $dom->loadHTML($cleaned_content);

      // Structure will be wrapped in <html><body> so we go 2 level deep.
      $childNodes = $dom->childNodes[1]->childNodes[0]->childNodes;
      if ($childNodes->length > 1) {
        $wrap = urlencode($cleaned_content);
      }
      libxml_use_internal_errors(FALSE);
    }

    return $wrap;
  }

  /**
   * For profiler to be active special cookie is required.
   */
  protected function profilingIsActive(): bool {
    /** @var \Drupal\qtools_cache_profiler\Cache\CacheProfilerActive $profilerCacheContext */
    $profilerCacheContext = \Drupal::service('cache_context.qtools_cache_profiler_active');
    return $profilerCacheContext->getContext();
  }

}
